package algorithms;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import nanoxml.XMLElement;
import util.AscendingOrderMyNodeComparator;
import util.EditPath;
import util.GRECCostFunction1;
import util.Graph;
import util.ICostFunction;
import util.IEdgeHandler;
import util.MutagenCostFunction1;
import util.MyTree;
import util.UniversalEdgeHandler;
import util.MyTree.MyNode;
import util.Node;
import xml.XMLParser;

public class GEDDFSMUTA {

	//double wctAvant;  
	MyTree OPEN; // Tiens  jour la liste des noeud non encore utiliss. Le score du chemin.

	private Graph G1; //Graph 1	
	private Graph G2; //Graph 2
	private Node CurVertex; //Current node
	private int G2NbNodes; //Number of nodes of G2


	boolean debug;

	private double UBCOST;
	public EditPath UB;
	private int heuristicmethod;
	private AscendingOrderMyNodeComparator MyNodeComparator;
	private int ubmethod;
	

	


	////////////////////////////////////////////////////////////////////////////////////////////
	public GEDDFSMUTA(Graph G1, Graph G2,ICostFunction costfunction, IEdgeHandler edgehandler,int heuristic, int ubmethod,boolean variableordering,boolean debug)
	{
		inittimer();
		
		this.debug = debug;
		this.G1 =G1;
		this.G2=G2;
		G2NbNodes = G2.size(); // Number of nodes of G2

			
		UBCOST = Double.MAX_VALUE;
		heuristicmethod = heuristic;
		this.ubmethod = ubmethod;

		EditPath ROOT = new EditPath(G1,G2,this.heuristicmethod,variableordering);
		MyNodeComparator =new AscendingOrderMyNodeComparator(heuristicmethod);
		OPEN = new MyTree<EditPath>(ROOT,MyNodeComparator);

		//test si l'on a une premire borne sup calcul lors du prprocessing
		if(Constants.FirstUB==null || ubmethod!=MunkresUB){
			//avant la boucle
			UB = ComputeUpperBound(ubmethod);
		}else{
			UB = Constants.FirstUB;
			this.UBCOST=UB.getTotalCosts();
			Constants.FirstUB=null;
		}
				
		inti();


		UB = loop();
		
		task.cancel();
		Constants.timer.purge();
		Constants.edgecostmatrix=null;
		Constants.nodecostmatrix=null;
	}

	protected boolean timeconstraint=false;

	private Timer timer;

	private boolean issolutionoptimal=false;

	private TimerTask task;

	public static int MunkresUB=0;
	public static int DummyUB=1;

	public boolean isTimeconstraintover() {
		return timeconstraint;
	}

	private void inittimer() {
		// TODO Auto-generated method stub
		task = new TimerTask()
		{
			@Override
			public void run() 
			{
				timeconstraint=true;
			}	
		};
		
		//timer = new Timer();
		Constants.timer.scheduleAtFixedRate(task, Constants.timeconstraint, Constants.timeconstraint);
		//timer.scheduleAtFixedRate(task, 0, 100);
	}


	private EditPath ComputeUpperBound(int upperboundmethod) {
		// TODO Auto-generated method stub

		EditPath res=null;
		
		if(upperboundmethod ==-1){
			UBCOST = Double.MAX_VALUE;
			res = null;
		}
		
		if(upperboundmethod == MunkresUB){
			MatrixGenerator mgen = new MatrixGenerator();
			Munkres munkres = new Munkres();
			MunkresRec munkresRec = new MunkresRec();
			mgen.setMunkres(munkresRec);

			double[][] matrix = mgen.getMatrix(this.G1, G2);
			munkres.setGraphs(G1, G2);
			UBCOST = munkres.getCosts(matrix);
			res = munkres.ApproximatedEditPath();
			//EditPath ROOT = new EditPath(this.G1,this.G2,this.heuristicmethod); 
			//res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();

		}

		if(upperboundmethod == this.DummyUB){
			EditPath ROOT = new EditPath(this.G1,this.G2); 
			ROOT.setHeuristicmethod(this.heuristicmethod);
			res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();
		}
		return res;
	}

	private EditPath dummyUpperBound(EditPath p) {
		EditPath ptmp=null;
		if(p.isComplete() == false){
			ptmp = new EditPath(p);
			// TODO Auto-generated method stub
			int G1NbNodes = ptmp.getUnUsedNodes1().size();
			int G2NbNodes = ptmp.getUnUsedNodes2().size();
			// Substitution case
			//System.out.println("Substitution phase");
			for(int i=0;i<G1NbNodes  ;i++)
			{

				Node u = (Node) ptmp.getUnUsedNodes1().getFirst();
				if(i<G2NbNodes) {
					Node v = (Node) ptmp.getUnUsedNodes2().getFirst();
					ptmp.addDistortion(u, v);
				}

			}

			//Deletion case
			if(G1NbNodes > G2NbNodes)
			{
				//System.out.println("Deletion phase");
				int noOfDeletedNodes = G1NbNodes- G2NbNodes; 
				for(int i=0;i<noOfDeletedNodes;i++)
				{

					Node u = ptmp.getNext();
					ptmp.addDistortion( u , Constants.EPS_COMPONENT);

				}
			}

			//Insertion case
			else if(G1NbNodes < G2NbNodes)
			{
				//System.out.println("Insertion phase");
				int noOfInsertedNodes = G2NbNodes- G1NbNodes; 
				for(int i=0;i<noOfInsertedNodes;i++)
				{
					Node u = ptmp.getNextG2();
					ptmp.addDistortion( Constants.EPS_COMPONENT , u );	
				}
			}
		}else{
			return p;
		}

		return ptmp;
	}



	private EditPath loop() {

		
	
		
		boolean condition1;
		boolean condition2;

		if(OPEN.isEmpty() == true){
			issolutionoptimal=true;
			return UB;
		}

		//On cherche dans OPEN le chemin avec le cout minimum (Pmin) et on l'enlve de OPEN
		EditPath pmin=null;
		MyNode<EditPath> pminNode=null;
		MyNode<EditPath> CurNode=OPEN.root;
	
		while(true){
		
			pminNode=OPEN.pollFirstLowestCost(CurNode);	
			
			//condition1 : Test if there is any child to explore from CurNode
			//condition2 : Test if CurNode is ROOT node
			condition1 = (pminNode == null);
			condition2 = ((pminNode == null)&&(CurNode.parent != null));
			while( condition1 && condition2){
				//BackTrack : pminNode is null so CurNode has no children. so we get the parent of CurNode
				CurNode =OPEN.BackTrack(CurNode);
				//pollFirstLowestCost : get the cheapest Child Node. The node is removed from the tree
				pminNode=OPEN.pollFirstLowestCost(CurNode);	
				condition1 = (pminNode == null);
				condition2 = ((pminNode == null)&&(CurNode.parent != null));
			}
			//System.out.println("nbexlporednode="+nbexlporednode);		
			if(this.timeconstraint==true){
				//we have reached the time constraint
				return UB;
			}
			
			
			if((pminNode == null) && (CurNode.parent == null) ){
				//all the tree has been explored since pminNode is null and its CurNode is root.
				//no more children to explore
				issolutionoptimal=true;
				return UB;
			}

			pmin = pminNode.data;
		
			//We cut if pmin is higher than UBCOST
			if(pmin.getTotalCosts()+pmin.ComputeHeuristicCosts(heuristicmethod)<UBCOST){
				
				// generates all successors of node u in the search tree
				// and add them to open
								
				if(pmin.getUnUsedNodes1().size() > 0){
					this.CurVertex=pmin.getNext();
					if(debug==true) System.out.println("Current Node="+this.CurVertex.getId());

					//pour tous les noeuds (w)  de V2  qui ne sont pas encore utiliss dans Pmin. on ajoute dans pmin toutes les substitutions de uk+1 avec w.
					LinkedList<Node>UnUsedNodes2= pmin.getUnUsedNodes2();

					for(int i=0;i<UnUsedNodes2.size();i++){
						EditPath newpath = new EditPath(pmin);
						Node w = UnUsedNodes2.get(i);
						newpath.addDistortion(this.CurVertex, w);
						AddTreeNode(pminNode,newpath,this.UBCOST);
					}
					//On met dans pmin une suppresion de uk+1.
					EditPath newpath = new EditPath(pmin);
					newpath.addDistortion(this.CurVertex, Constants.EPS_COMPONENT);
					AddTreeNode(pminNode,newpath,this.UBCOST);


				}else{
					//Sinon si k= taille de G1
					//On met dans pmin toutes les insertions des noeuds  de V2  qui ne sont pas encore utiliss dans Pmin. 
					//On remet Pmin dans OPEN
					//System.out.println("nbexlporednode="+nbexlporednode);
					EditPath newpath = new EditPath(pmin);
					newpath.complete();	
					double g = newpath.getTotalCosts();
					double h = newpath.ComputeHeuristicCosts(heuristicmethod);
					double f= g+h;

					if(f<UBCOST)
					{
						UBCOST = f;
						UB = newpath;
						
					}
					
				}//end of if(pmin.getUnUsedNodes1().size() > 0)
			}//end of cut
			
			//next node to be explored
			//pminNode that has been expanded. Children has been created so we need to go deeper
			CurNode=pminNode;


		}//end of while true



	}


	// the initialization function
	private void inti() {
		// TODO Auto-generated method stub
		
		//System.out.println("Begin init");
		for(int i=0;i<G2NbNodes;i++){
			//Dans la liste OPEN on met toutes les substitions de u1 avec tous les noeuds de g2.
			//index du noeud courant

			EditPath p = new EditPath((EditPath) OPEN.root.data); // An intialization for both G1 and G2 "to show that we did not use neither edges nor nodes of both G1 and G2"
			Node v = (Node)G2.get(i);
			this.CurVertex = p.getNext();
			/* Add distortion between the current node u1 and each node in G2 
				 distortion means insertion or substitution but not deletion as we are sure that
				 G2 has nodes as we are inside the for loop of G2 */
			p.addDistortion(this.CurVertex, v);
		//	System.out.println("test ::"+p.toString());
		//	 System.out.println("Substitution CurNode G1 and ith node of G2= ("+this.CurVertex.getId()+"  "+v.getId()+")   Cost ="+p.getTotalCosts());
			AddTreeNode(OPEN.root,p,this.UBCOST);


		}

		//On met dans OPEN une suppresion de u1. Dans le cas ou supprimer u1 serait une possibilit optimale.
		EditPath p = new EditPath((EditPath) OPEN.root.data);
		this.CurVertex = p.getNext();
		p.addDistortion(this.CurVertex, Constants.EPS_COMPONENT); // Zeina: deletion
	//	System.out.println("test ::"+p.toString());

		if(debug==true) System.out.println("Deletion CurNode="+this.CurVertex.getId() + " -- Cost = "+p.getTotalCosts());
		AddTreeNode(OPEN.root,p,this.UBCOST);
		
		//System.out.println("Finish init ....");

	}

//add a node to the tree if it is worth. if it is lower than UBCOST
//the new node is linked to its parent node.
	private void AddTreeNode(MyNode<EditPath> parent, EditPath p, double ubcost2) {
		// TODO Auto-generated method stub
		double g=p.getTotalCosts();
		double h=p.ComputeHeuristicCosts(heuristicmethod);
		double f = g+h;
		if(f<ubcost2)
		{
			OPEN.Add(parent,p);	
		}
	}


	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		// TODO Auto-generated method stub
		Graph g2,g1;
		
		Constants.timer = new Timer();	
		Constants.timeconstraint = 30000;  // time limit: 30 seconds


		XMLParser xmlParser = new XMLParser();

		
		double nodesub = Double.parseDouble(args[0]);
		double nodeinsertdel = Double.parseDouble(args[1]);
		
		double edgesub = Double.parseDouble(args[2]);
		double edgeinsertdel = Double.parseDouble(args[3]);

		
		g1=xmlParser.parseGXL(args[4]);
		g2=xmlParser.parseGXL(args[5]);
		
		

		
		Constants.costFunction = new MutagenCostFunction1(nodesub,nodeinsertdel,edgesub,edgeinsertdel);
		Constants.edgeHandler = new UniversalEdgeHandler(); 

		long start = System.nanoTime();

		GEDDFSMUTA HGED = new GEDDFSMUTA(g1,g2,Constants.costFunction,Constants.edgeHandler,EditPath.NoAssigmentHeuristic,0,false,false);
		long end =  System.nanoTime();
		
		//System.out.println("Distance ::: "+HGED.getBestEditpath().getTotalCosts());
		//System.out.println("Distance ::: "+HGED.isTimeconstraintover());

		

		long executionTime_NanoSeconds = end - start;
				
		double executionTime_seconds = (double)executionTime_NanoSeconds / (double)1000000000;
		
		

		
		
		//ps.println("Opt_sol;time;node_match");
		System.out.print(HGED.issolutionoptimal+";");
		System.out.print(new DecimalFormat("##.########").format(executionTime_seconds)+";");
		

		System.out.print(HGED.getBestEditpath().g2Indices1(g1));
		System.out.println(";"+HGED.getBestEditpath().getTotalCosts());
		System.exit(0);	
	}



	private static FilenameFilter gxlFileFilter = new FilenameFilter() {
		public boolean accept(File dir, String name) {
			return name.endsWith(".gxl");
		}
	};



	public EditPath getBestEditpath() {
		// TODO Auto-generated method stub
			return UB;		
	}
	


	public boolean isSolutionoptimal() {
		// TODO Auto-generated method stub
		return issolutionoptimal;
	}
	
	private boolean memoryconstraint=false;
	
	public boolean isMemoryconstraintover() {
		return memoryconstraint;
	}

	
	
	
}
