package algorithms;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.ReentrantLock;

import algorithms.Constants;
import algorithms.GEDDFSGREC;
import algorithms.MatrixGenerator;
import algorithms.Munkres;
import algorithms.MunkresRec;

import nanoxml.XMLElement;


import algorithms.AscendingOrderNonStaticNodeComparator;
import algorithms.HeuristicGEDPartialMatcher;
import util.AscendingHeuristicEditPathComparator;
import util.CMUHouseCostFunction1;
import util.EditPath;
import util.GRECCostFunction1;
import util.Graph;
import util.ICostFunction;
import util.IEdgeHandler;
import util.MutagenCostFunction1;
import util.MyTree;
import util.UniversalEdgeHandler;
import util.MyTree.MyNode;
import util.Node;
import xml.XMLParser;

public class GEDMultiThreadCMU {

	//double wctAvant;  
	MyTree OPEN; // Tiens  jour la liste des noeud non encore utiliss. Le score du chemin.

	private Graph G1; //Graph 1	
	private Graph G2; //Graph 2
	private Node CurVertex; //Current node
	private int G2NbNodes; //Number of nodes of G2
	


	boolean debug;

	private double UBCOST;


	private int heuristicmethod;


	private AscendingOrderNonStaticNodeComparator MyNodeComparator;

	private int ubmethod;

	

	



	////////////////////////////////////////////////////////////////////////////////////////////
	public GEDMultiThreadCMU(Graph G1, Graph G2,ICostFunction costfunction, IEdgeHandler edgehandler,int heuristic, int ubmethod,boolean variableordering,int nbthread,int numberofInitialEditPaths,boolean debug){
		//start the timer
		//System.out.println("/////////////////////////////////////////////////////////////////////");

		inittimer();
		//set some variable to default value
		GlobalVar.selectedthread=-1;
		GlobalVar.lightthread=-1;
		GlobalVar.timeconstraint=false;
		GlobalVar.memoryconstraint=false;
		GlobalVar.averageworkload=0;
		GlobalVar.tabthreads=null;
		GlobalVar.UBCOST=Double.MAX_VALUE;
		GlobalVar.UB=null;
		this.debug = debug;
		this.G1 =G1;
		this.G2=G2;
		G2NbNodes = G2.size(); // Number of nodes of G2

		UBCOST = Double.MAX_VALUE;
		heuristicmethod = heuristic;
		this.ubmethod = ubmethod;
			
		
		if(numberofInitialEditPaths<G1.size() || numberofInitialEditPaths<G2.size()){
			numberofInitialEditPaths = Math.max(G1.size(), G2.size())+1;

		}
				
		
		//we set up the semaphore to - nbthread +1
		//the goal is to make the main thread wait until all the thread have finished
		GlobalVar.sem = new Semaphore(-nbthread+1);
		//the mutex ready to block when the load balancing will be needed
		GlobalVar.mutex = new Semaphore(0);
			
		//first tree to decompose the load
		//EditPath ROOT = new EditPath(G1,G2,this.heuristicmethod,variableordering);
		//each thread has its own local tree search with a root node			
		MyNodeComparator =new AscendingOrderNonStaticNodeComparator(heuristicmethod);	
		//OPEN = new MyTree<EditPath>(ROOT,MyNodeComparator);
		
		EditPath ROOT1 = new EditPath(G1,G2,this.heuristicmethod,variableordering);

		HeuristicGEDPartialMatcher HGED =  new HeuristicGEDPartialMatcher(G1, G2, numberofInitialEditPaths ,Constants.costFunction, Constants.edgeHandler,heuristicmethod,ROOT1,ubmethod,variableordering,debug);
		AscendingHeuristicEditPathComparator EDComparator = new AscendingHeuristicEditPathComparator(heuristicmethod);
		Collections.sort(HGED.OPEN,EDComparator);
		ArrayList<EditPath> somejobs= HGED.OPEN;
		
		
		//the table of thread is created
		GlobalVar.tabthreads = new GEDDFSThread[nbthread];
		

		GlobalVar.threadNoOfIterations = new double[nbthread];
		GlobalVar.threadvariance = new double[nbthread];

		
		//initialization of thread
		for(int i=0;i<nbthread;i++){
			
			
			// instanciation of threads
			GlobalVar.threadNoOfIterations[i]=0.0;
			GlobalVar.threadvariance[i] = 0.0;
			
			
			GlobalVar.tabthreads[i]= new GEDDFSThread(heuristicmethod,debug);
			
			//set the root node
			GlobalVar.tabthreads[i].SetRootNode(ROOT1, MyNodeComparator);
			//set the id of the thread
			GlobalVar.tabthreads[i].SetIdThead(i);
		}
		//we dispath the jobs amaong all the thread
		//using modulo
		DispatchJobs(GlobalVar.tabthreads,somejobs);
		//we start all the threads
		for(int i=0;i<nbthread;i++){
			GlobalVar.tabthreads[i].start();		
		}
		
		//a small sleep cause a thread could have already release the semaphore
		try {
			Thread.currentThread().sleep(1);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		//we block here waiting for  all the threads to finish
		try {
			GlobalVar.sem.acquire();
		} catch (InterruptedException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	
		
		//we finish the timer
		task.cancel();
		Constants.timer.purge();
		//we clean variables
		Constants.edgecostmatrix=null;
		Constants.nodecostmatrix=null;
	}
	
	

	private void DispatchJobs(GEDDFSThread[] tabthreads,
			ArrayList<EditPath> somejobs) {

		
	//	System.out.println("Dispatching work .........");
		for(int i=0;i<somejobs.size();i++){
			int index = i%tabthreads.length;
			tabthreads[index].addJob(somejobs.get(i));
		}
	}

	//protected boolean timeconstraint=false;

	private Timer timer;

	private boolean issolutionoptimal=false;

	private TimerTask task;

	public static int MunkresUB=0;
	public static int DummyUB=1;

	public boolean isTimeconstraintover() {
		return GlobalVar.timeconstraint;
	}

	private void inittimer() {
		// TODO Auto-generated method stub
		task = new TimerTask()
		{
			@Override
			public void run() 
			{
				GlobalVar.timeconstraint=true;
			}	
		};
		
		//timer = new Timer();
		Constants.timer.scheduleAtFixedRate(task, Constants.timeconstraint, Constants.timeconstraint);
		//timer.scheduleAtFixedRate(task, 0, 100);
	}


	private EditPath ComputeUpperBound(int upperboundmethod) {
		// TODO Auto-generated method stub

		EditPath res=null;
		
		if(upperboundmethod ==-1){
			UBCOST = Double.MAX_VALUE;
			res = null;
		}
		
		if(upperboundmethod == MunkresUB){
			MatrixGenerator mgen = new MatrixGenerator();
			Munkres munkres = new Munkres();
			MunkresRec munkresRec = new MunkresRec();
			mgen.setMunkres(munkresRec);

			double[][] matrix = mgen.getMatrix(this.G1, G2);
			munkres.setGraphs(G1, G2);
			UBCOST = munkres.getCosts(matrix);
			res = munkres.ApproximatedEditPath();
			//EditPath ROOT = new EditPath(this.G1,this.G2,this.heuristicmethod); 
			//res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();

		}

		if(upperboundmethod == this.DummyUB){
			EditPath ROOT = new EditPath(this.G1,this.G2,this.heuristicmethod,false); 
			res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();

		}
		return res;
	}

	private EditPath dummyUpperBound(EditPath p) {
		EditPath ptmp=null;
		if(p.isComplete() == false){
			ptmp = new EditPath(p);
			// TODO Auto-generated method stub
			int G1NbNodes = ptmp.getUnUsedNodes1().size();
			int G2NbNodes = ptmp.getUnUsedNodes2().size();
			// Substitution case
			//System.out.println("Substitution phase");
			for(int i=0;i<G1NbNodes  ;i++)
			{

				Node u = (Node) ptmp.getUnUsedNodes1().getFirst();
				if(i<G2NbNodes) {
					Node v = (Node) ptmp.getUnUsedNodes2().getFirst();
					ptmp.addDistortion(u, v);
				}

			}

			//Deletion case
			if(G1NbNodes > G2NbNodes)
			{
				//System.out.println("Deletion phase");
				int noOfDeletedNodes = G1NbNodes- G2NbNodes; 
				for(int i=0;i<noOfDeletedNodes;i++)
				{

					Node u = ptmp.getNext();
					ptmp.addDistortion( u , Constants.EPS_COMPONENT);

				}
			}

			//Insertion case
			else if(G1NbNodes < G2NbNodes)
			{
				//System.out.println("Insertion phase");
				int noOfInsertedNodes = G2NbNodes- G1NbNodes; 
				for(int i=0;i<noOfInsertedNodes;i++)
				{
					Node u = ptmp.getNextG2();
					ptmp.addDistortion( Constants.EPS_COMPONENT , u );	
				}
			}
		}else{
			return p;
		}

		return ptmp;
	}



	

	// the initialization function
	private ArrayList<EditPath> inti() {
		// TODO Auto-generated method stub
		 ArrayList<EditPath> res = new  ArrayList<EditPath>();
		for(int i=0;i<G2NbNodes;i++){
			//Dans la liste OPEN on met toutes les substitions de u1 avec tous les noeuds de g2.
			//index du noeud courant

			EditPath p = new EditPath((EditPath) OPEN.root.data); // An intialization for both G1 and G2 "to show that we did not use neither edges nor nodes of both G1 and G2"
			Node v = (Node)G2.get(i);
			this.CurVertex = p.getNext();
			/* Add distortion between the current node u1 and each node in G2 
				 distortion means insertion or substitution but not deletion as we are sure that
				 G2 has nodes as we are inside the for loop of G2 */
			p.addDistortion(this.CurVertex, v);
			if(debug==true) System.out.println("Substitution CurNode G1 and ith node of G2= ("+this.CurVertex.getId()+"  "+v.getId()+")   Cost ="+p.getTotalCosts());
			res.add(p);
			AddTreeNode(OPEN.root,p,this.UBCOST);



		}

		//On met dans OPEN une suppresion de u1. Dans le cas ou supprimer u1 serait une possibilit optimale.
		EditPath p = new EditPath((EditPath) OPEN.root.data);
		this.CurVertex = p.getNext();
		p.addDistortion(this.CurVertex, Constants.EPS_COMPONENT); // Zeina: deletion
		if(debug==true) System.out.println("Deletion CurNode="+this.CurVertex.getId() + " -- Cost = "+p.getTotalCosts());
		res.add(p);
		AddTreeNode(OPEN.root,p,this.UBCOST);
		
		return res;

	}

//add a node to the tree if it is worth. if it is lower than UBCOST
//the new node is linked to its parent node.
	private void AddTreeNode(MyNode<EditPath> parent, EditPath p, double ubcost2) {
		// TODO Auto-generated method stub
		double g=p.getTotalCosts();
		double h=p.ComputeHeuristicCosts(heuristicmethod);
		double f = g+h;
		if(f<ubcost2){
			OPEN.Add(parent,p);	
		}
	}


	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		// TODO Auto-generated method stub
		Graph g2,g1;
		
		Constants.timer = new Timer();	
				
		Constants.timeconstraint = 30000;  // time limit: 30 seconds
		Constants.costFunction = new CMUHouseCostFunction1();
		Constants.edgeHandler = new UniversalEdgeHandler(); 
				

		XMLParser xmlParser = new XMLParser();

		g1=xmlParser.parseGXL(args[0]);
		g2=xmlParser.parseGXL(args[1]);
				
		long start = System.nanoTime();
	
		GEDMultiThreadCMU HGED = new GEDMultiThreadCMU(g1,g2,Constants.costFunction,Constants.edgeHandler,EditPath.MunkresAssigmentHeuristic,MunkresUB,true,128,-1,false);

		long end = System.nanoTime();
		System.out.println("Distance ::: "+HGED.getBestEditpath().getTotalCosts());
		System.out.println("Distance ::: "+HGED.isTimeconstraintover());
		
		

		long executionTime_NanoSeconds = end - start;
				
		double executionTime_seconds = (double)executionTime_NanoSeconds / (double)1000000000;
		
		

		
		PrintStream ps = new PrintStream(args[2]);
		
		ps.println("Opt_sol;time;node_match");
		ps.print(HGED.issolutionoptimal+";");
		ps.print(new DecimalFormat("##.########").format(executionTime_seconds)+";");
		

		ps.println(HGED.getBestEditpath().g2Indices1(g1));
		ps.close();

		System.exit(0);

	}

	
	

	private static FilenameFilter gxlFileFilter = new FilenameFilter() {
		public boolean accept(File dir, String name) {
			return name.endsWith(".gxl");
		}
	};





	public EditPath getBestEditpath() {
		// TODO Auto-generated method stub
		return GlobalVar.UB;
	}
	
	
	///////////////////////////////////////////////////////////////////////////////////////////////////
	public double getVariance() {
		// TODO Auto-generated method stub

		double maxNoOfIterations = -1;
		int threadIndex = -1;
		for(int i=0;i<GlobalVar.tabthreads.length;i++)
		{
			if(GlobalVar.threadNoOfIterations[i]>maxNoOfIterations)
			{
				maxNoOfIterations = GlobalVar.threadNoOfIterations[i];
				threadIndex = i;
			}
		}
		
		double averageVariance = GlobalVar.threadvariance[threadIndex]/maxNoOfIterations;
		
		return averageVariance;
		
	}
	
	public double getMaxIteration() {
		// TODO Auto-generated method stub

		double maxNoOfIterations = -1;
		int threadIndex = -1;
		for(int i=0;i<GlobalVar.tabthreads.length;i++)
		{
			if(GlobalVar.threadNoOfIterations[i]>maxNoOfIterations)
			{
				maxNoOfIterations = GlobalVar.threadNoOfIterations[i];
				threadIndex = i;
			}
		}
		
		
		return maxNoOfIterations;
		
	}
	///////////////////////////////////////////////////////////////////////////////////////////////////

	
	public double getIDLEtime() {
		// TODO Auto-generated method stub
		//double maxcputime = GlobalVar.tabthreads.length*Constants.timeconstraint;
		double res=0;
		for(int i=0;i<GlobalVar.tabthreads.length;i++){
			res+=GlobalVar.tabthreads[i].getIdletime();
		}
		//return maxcputime-res;
		return res;
	}
	
	public int getNbExploredNode() {
		// TODO Auto-generated method stub
		int res=0;
		for(int i=0;i<GlobalVar.tabthreads.length;i++){
			res+=GlobalVar.tabthreads[i].getNbExploredNode();
		}
		return res;
		
	}

	public double getCputime() {
		double res=0;
		for(int i=0;i<GlobalVar.tabthreads.length;i++){
			res+=GlobalVar.tabthreads[i].getCputime();
		}
		return res;
	}
	
	public int getMaxSizeOpen() {
		int res=0;
		for(int i=0;i<GlobalVar.tabthreads.length;i++){
			res+=GlobalVar.tabthreads[i].getMaxSizeOpen();
		}
		return res;
	}

	public boolean isSolutionoptimal() {
		// TODO Auto-generated method stub
		
//		for(int i=0;i<GlobalVar.tabthreads.length;i++){
//			if(GlobalVar.tabthreads[i].isSolutionoptimal()==false){
//				issolutionoptimal=false;
//			}
//		}	
//		return true;
		
	     if(isTimeconstraintover()==true)
         {
                 return false;
         }
         else
         {
                 return true;

         }

	}
	
	


	
	
	
}
