package algorithms;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;

import algorithms.Constants;
import algorithms.MatrixGenerator;
import algorithms.Munkres;
import algorithms.MunkresRec;


import util.EditPath;

import util.Graph;
import util.ICostFunction;
import util.IEdgeHandler;
import util.Node;
import util.UniversalEdgeHandler;
import util.UnlabeledCostFunction;
import xml.XMLParser;
import xml.XMLWriter;

/**
 * @author Romain Raveaux
 * 
 */

public class HeuristicGEDPartialMatcher {


	public static int MunkresUB=0;
	public static int DummyUB=1;

	public ArrayList<EditPath> OPEN; //Tiens � jour la liste des noeud non encore utilis�s. Le score du chemin.
	private Graph G1; // Graph 1	
	private Graph G2; // Graph 2
	private Node CurNode; //Current node
	private int G2NbNodes; //number of nodes of G2
	public int openCounterSize; // a variable that tracks the maximum size of the set OPEN
	public  int editPathCounter; // a variable that tracks number of edit paths
	public int noOfEditPaths;
	boolean v1Sorting;
	
	public double UBCOST=Double.MAX_VALUE;
	public EditPath UBEditPath;

	//debug info
		//debug variable
		int nbexlporednode;
		int maxopensize=0;
	
	private EditPath BestEditpath=null; // Zeina: The best answer
	boolean debug;
	private int heuristicmethod;
	private boolean isthereanyfeasibleolutionfound=false;
	EditPath root;

	public boolean isIsthereanyfeasibleolutionfound() {
		return isthereanyfeasibleolutionfound;
	}

	// A function that returns the best edit path "the optimal solution"
	public EditPath getBestEditpath() {
		return BestEditpath;
	}
	
   public HeuristicGEDPartialMatcher(Graph G1, Graph G2, int noOfEditPaths,ICostFunction costfunction, IEdgeHandler edgehandler,int heuristicmethod,boolean v1sorting,boolean debug,  EditPath upperBound){
		
	   	this.v1Sorting = v1sorting;
		this.debug = debug;
		this.G1 =G1;
		this.G2=G2;
		this.heuristicmethod=heuristicmethod;
		this.noOfEditPaths = noOfEditPaths;
		G2NbNodes = G2.size(); // Number of nodes of G2
	//	BestEditpath=null; // In the beginning, the list is empty. 
		//The set open contains the search tree nodes to be processed in the next steps
		OPEN = new ArrayList<EditPath>();
		
		UBCOST = upperBound.getTotalCosts();
		UBEditPath = new EditPath(upperBound);

		//avant la boucle
		inti();
		nbexlporednode++;

		BestEditpath = loop();
		
	/*	timer.cancel();
		timer.purge();
		*/
		//Constants.edgecostmatrix=null;
		//Constants.nodecostmatrix=null;
		
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
   public HeuristicGEDPartialMatcher(Graph G1, Graph G2, int noOfEditPaths,ICostFunction costfunction, IEdgeHandler edgehandler,int heuristicmethod,EditPath root,int ubmethod,boolean v1sorting,boolean debug){
		this.root=root;
	   	this.v1Sorting = v1sorting;
		this.debug = debug;
		this.G1 =G1;
		this.G2=G2;
		this.heuristicmethod=heuristicmethod;
		this.noOfEditPaths = noOfEditPaths;
		G2NbNodes = G2.size(); // Number of nodes of G2
	//	BestEditpath=null; // In the beginning, the list is empty. 
		//The set open contains the search tree nodes to be processed in the next steps
		OPEN = new ArrayList<EditPath>();
		
		//test si l'on a une premire borne sup calcul lors du prprocessing
		if(Constants.FirstUB==null || ubmethod!=MunkresUB){
			//avant la boucle
			GlobalVar.UB = ComputeUpperBound(ubmethod);
			if(ubmethod==-1)
			{
				GlobalVar.UBCOST = Double.MAX_VALUE;
				this.UBCOST=Double.MAX_VALUE;
			}
			else
			{
				GlobalVar.UBCOST = GlobalVar.UB.getTotalCosts();
				this.UBCOST=GlobalVar.UB .getTotalCosts();
			}
			
		}else{
			GlobalVar.UB = Constants.FirstUB;
			GlobalVar.UBCOST = GlobalVar.UB.getTotalCosts();

			this.UBCOST=GlobalVar.UB .getTotalCosts();
			Constants.FirstUB=null;
		}
		//avant la boucle
		inti();
		nbexlporednode++;

		BestEditpath = loop();

		
	}
	
	
   private EditPath ComputeUpperBound(int upperboundmethod) {
		// TODO Auto-generated method stub

		EditPath res=null;
		
		if(upperboundmethod ==-1){
			UBCOST = Double.MAX_VALUE;
			res = null;
		}
		
		if(upperboundmethod == MunkresUB){
			MatrixGenerator mgen = new MatrixGenerator();
			Munkres munkres = new Munkres();
			MunkresRec munkresRec = new MunkresRec();
			mgen.setMunkres(munkresRec);

			double[][] matrix = mgen.getMatrix(this.G1, G2);
			munkres.setGraphs(G1, G2);
			UBCOST = munkres.getCosts(matrix);
			res = munkres.ApproximatedEditPath();
			//EditPath ROOT = new EditPath(this.G1,this.G2,this.heuristicmethod); 
			//res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();

		}

		if(upperboundmethod == this.DummyUB){
			EditPath ROOT = new EditPath(this.G1,this.G2,this.heuristicmethod,false); 
			res = this.dummyUpperBound(ROOT);
			UBCOST = res.getTotalCosts();

		}
		return res;
	}

	private EditPath dummyUpperBound(EditPath p) {
		EditPath ptmp=null;
		if(p.isComplete() == false){
			ptmp = new EditPath(p);
			// TODO Auto-generated method stub
			int G1NbNodes = ptmp.getUnUsedNodes1().size();
			int G2NbNodes = ptmp.getUnUsedNodes2().size();
			// Substitution case
			//System.out.println("Substitution phase");
			for(int i=0;i<G1NbNodes  ;i++)
			{

				Node u = (Node) ptmp.getUnUsedNodes1().getFirst();
				if(i<G2NbNodes) {
					Node v = (Node) ptmp.getUnUsedNodes2().getFirst();
					ptmp.addDistortion(u, v);
				}

			}

			//Deletion case
			if(G1NbNodes > G2NbNodes)
			{
				//System.out.println("Deletion phase");
				int noOfDeletedNodes = G1NbNodes- G2NbNodes; 
				for(int i=0;i<noOfDeletedNodes;i++)
				{

					Node u = ptmp.getNext();
					ptmp.addDistortion( u , Constants.EPS_COMPONENT);

				}
			}

			//Insertion case
			else if(G1NbNodes < G2NbNodes)
			{
				//System.out.println("Insertion phase");
				int noOfInsertedNodes = G2NbNodes- G1NbNodes; 
				for(int i=0;i<noOfInsertedNodes;i++)
				{
					Node u = ptmp.getNextG2();
					ptmp.addDistortion( Constants.EPS_COMPONENT , u );	
				}
			}
		}else{
			return p;
		}

		return ptmp;
	}


	
	///////////////////////////////////////////////////////////////////////////////////////////
	private EditPath loop() {
		// TODO Auto-generated method stub

		
		while(true == true){
			
			if(OPEN.isEmpty() == true){
			//	System.out.println("empty ...");
			//	System.out.println("Error : No more candidates, no complete solution could be found");
			//	System.out.println("Error : Please check your graph data");
				return null;
			}
			if(debug==true) System.out.println("----------------------------------------------------------");
			
			if(debug==true)System.out.println("SIZE  ...."+this.OPEN.size()+" no of edit paths "+noOfEditPaths);
			if(OPEN.size() > noOfEditPaths)
			{
			//	System.out.println("this size of OPEN is "+OPEN.size() );
				return null;
			}
			
			//On cherche dans OPEN le chemin avec le cout minimum (Pmin) et on l'enl�ve de OPEN
			EditPath pmin =findmincostEditPathWithHeuristic(OPEN);
			
			this.nbexlporednode++;
			if(debug==true) System.out.println("nb explored nodes"+nbexlporednode);
			OPEN.remove(pmin);
			
			if(debug==true)System.out.println("-------------------------------------------------------------------------------------");

			if(debug==true)System.out.println("Choosing "+pmin.toString());
			

			if(pmin.isComplete() == true){
		
				UBCOST = pmin.getTotalCosts();
				UBEditPath = new EditPath(pmin);
				GlobalVar.UBCOST = pmin.getTotalCosts();
				GlobalVar.UB = new EditPath(pmin);
				OPEN.clear();
				return pmin;
			}else{
	
				
				if(pmin.getUnUsedNodes1().size() > 0){;
					this.CurNode=pmin.getNext();
					
					// pour tous les noeuds (w)  de V2  qui ne sont pas encore utilis�s dans Pmin. on ajoute dans pmin toutes les substitutions de uk+1 avec w.
					LinkedList<Node>UnUsedNodes2= pmin.getUnUsedNodes2();
					for(int i=0;i<UnUsedNodes2.size();i++){
						EditPath newpath = new EditPath(pmin);
						
						Node w = UnUsedNodes2.get(i);
						newpath.addDistortion(this.CurNode, w);
						if(debug==true)System.out.println("***Substitution ");


						if(newpath.getTotalCosts()+newpath.ComputeHeuristicCosts(this.heuristicmethod)<UBCOST)
						{
						     this.OPEN.add(newpath);
						     if(debug==true) System.out.println("Substitution "+newpath.toString());

						}
					//	if(debug==true) TrackOpenEditPathSize();
					//	if(debug==true) System.out.println("----------------------------------------------------------");
					}
					// On met dans pmin une suppresion de uk+1.
					EditPath newpath = new EditPath(pmin);
					newpath.addDistortion(this.CurNode, Constants.EPS_COMPONENT);
				
		
					if(debug==true) System.out.println("****deletion case ");
					if(newpath.getTotalCosts()+newpath.ComputeHeuristicCosts(this.heuristicmethod)<UBCOST)
					{
				     	this.OPEN.add(newpath);
				     	if(debug==true) System.out.println("Deletion "+newpath.toString());
					}
					if(debug==true) TrackOpenEditPathSize();

				}else{
					//Sinon si k= taille de G1
					//On met dans pmin toutes les insertions des noeuds  de V2  qui ne sont pas encore utilis�s dans Pmin. 
					//On remet Pmin dans OPEN
					EditPath newpath = new EditPath(pmin);
					newpath.complete();	
					
					// added by Zeina for setup time 
				//	GlobalVar.timeconstraint = true;
					//this.OPEN.add(newpath);
					
					if(debug==true) System.out.println("**Trying to update UB from"+ UBCOST+ "to." +newpath.getTotalCosts()+ "");
					if(newpath.getTotalCosts()+newpath.ComputeHeuristicCosts(this.heuristicmethod)<UBCOST)
					{
						UBCOST = newpath.getTotalCosts();
						UBEditPath = newpath;//new EditPath(newpath);
						GlobalVar.UBCOST=UBCOST;
						GlobalVar.UB=newpath;//UBEditPath;
						if(debug==true) System.out.println("modifying upper bound"+UBCOST);
					}
					

					
					if(debug==true) TrackOpenEditPathSize();
					
					if(debug==true)
					{
						System.out.println("We complete the path by inserting all remaining nodes of G2 CurNode="+this.CurNode.getId());
						newpath.printMe();
					}
						
				}//fin si k<(this.G1NbNodes-1)
			}//fin si chemin optimal

		}//boucle while
	}
	
	

	//A function that returns the minimal edit path's cost in the set OPEN
	private EditPath findmincostEditPathWithHeuristic(ArrayList<EditPath> oPEN2) {
		// TODO Auto-generated method stub
		int i=0;
		int nbpaths = OPEN.size();
		double minvalue = Double.MAX_VALUE;
		double h=0.0,g=0.0;
		int indexmin = -1;

		for(i=0;i<nbpaths;i++){
			EditPath p = OPEN.get(i);
			g = p.getTotalCosts(); // computing the cost of the edit path from the root till the current node...
			h=p.ComputeHeuristicCosts(this.heuristicmethod);

			
			
			if((g+h) <minvalue){
				minvalue = g+h;
				indexmin = i;
			}
		}

		
		return OPEN.get(indexmin);
	}
	
	//----------------------------------------------------------------------------------------
	// the initialization function
	private void inti() {
		
		if(G1.size() ==0 && G2.size() ==0){
			System.out.println("G1 has no node inside");
			System.out.println("G2 has no node inside");
			System.out.println("I cannot work !!!");
		}
		
		if(G1.size() ==0 ){
			System.out.println("G1 has no node inside");
			this.G1=G2;
			this.G2=G1;
			G2NbNodes = G2.size();
			//p = new EditPath(G2,G1,this.heuristicmethod);
		}
		
		//O(n^3)
		//root = new EditPath(G1,G2,this.heuristicmethod, v1Sorting); // An intialization for both G1 and G2 "to show that we did not use neither edges nor nodes of both G1 and G2"
		
		// TODO Auto-generated method stub
		for(int i=0;i<G2NbNodes;i++){
			//Dans la liste OPEN on met toutes les substitions de u1 avec tous les noeuds de g2.
			//index du noeud courant
			EditPath p = new EditPath(root); // An intialization for both G1 and G2 "to show that we did not use neither edges nor nodes of both G1 and G2"
		
			Node v = (Node)G2.get(i);
			this.CurNode = p.getNext();
			/*Add distortion between the current node u1 and each node in G2 
			 distortion means insertion or substitution but not deletion as we are sure that
			 G2 has nodes as we are inside the for loop of G2 */
			p.addDistortion(this.CurNode, v);
			if(debug==true) System.out.println("Substitution "+p.toString());
			OPEN.add(p);
			//if(debug==true) TrackOpenEditPathSize();
		}

		//L3 : On met dans OPEN une suppresion de u1. Dans le cas ou supprimer u1 serait une possibilit� optimale.
		EditPath p = new EditPath(root);
		this.CurNode = p.getNext();
		p.addDistortion(this.CurNode, Constants.EPS_COMPONENT); // Zeina: deletion
		if(debug==true) System.out.println("Deletion "+p.toString());
		OPEN.add(p);
		//if(debug==true) TrackOpenEditPathSize();
	}
	//----------------------------------------------------------------------------------------

	
	// A function that tracks the no of unexamined "added" edit Paths and the maximum size of the set OPEN
	private void TrackOpenEditPathSize() {
		// TODO Auto-generated method stub
		this.editPathCounter++;
		if(openCounterSize < OPEN.size())
		{
			openCounterSize=OPEN.size();
		/*	if(debug==true)
			{
				System.out.println("OPEN max size = "+openCounterSize);
				System.out.println("No of edit paths = "+editPathCounter);
			
			}
		*/	
		}
	}

	
	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		
		System.out.println("Before executing the Heuristic Graph Edit Distance function");

		
		System.out.println("After executing the Heuristic Graph Edit Distance function ");
		
	}
	
	private static FilenameFilter gxlFileFilter = new FilenameFilter() {
		public boolean accept(File dir, String name) {
			return name.endsWith(".gxl");
		}
	};
	
	

	public int getNbExploredNode() {
		// TODO Auto-generated method stub
		return this.nbexlporednode;
	}

	public int getMaxSizeOpen() {
		// TODO Auto-generated method stub
		return this.maxopensize;
	}
	
	
	protected boolean timeconstraint=false;
	private Timer timer;

	public boolean isTimeconstraintover() {
		return timeconstraint;
	}

	private void inittimer() {
		// TODO Auto-generated method stub
		TimerTask task = new TimerTask()
		{
			@Override
			public void run() 
			{
				timeconstraint=true;
			}	
		};
		
		timer = new Timer();
		timer.scheduleAtFixedRate(task, 60000, 60000);
	}
	
	private boolean memoryconstraint=false;
	private boolean issolutionoptimal=false;

	public boolean isMemoryconstraintover() {
		return memoryconstraint;
	}

	private boolean memorylimittest() {
		// TODO Auto-generated method stub
		long val =  Runtime.getRuntime().freeMemory();
		if(val<1000000){
			memoryconstraint=true;
			return true;
		}
		return false;
	}

	public boolean isSolutionoptimal() {
		// TODO Auto-generated method stub
		return issolutionoptimal;
	}
}
