/**
 * 
 */
package algorithms;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.Properties;
import java.util.Timer;

import util.CMUHouseCostFunction1;
import util.GRECCostFunction1;
import util.Graph;
import util.GraphCollection;
import util.UniversalEdgeHandler;
import xml.XMLParser;
import xml.XMLWriter;

/**
 * @author kriesen
 * 
 */
public class MainMunkresGREC {

	/**
	 * the collections
	 */
	private GraphCollection source, target;

	/**
	 * the results of all matchings
	 */
	private double[][] distanceMatrix;

	/**
	 * the source and target graphs
	 */
	private Graph sourceGraph, targetGraph;

	/**
	 * boolean to check if swapped or not
	 */
	private boolean swapped = false;

	/**
	 * 
	 */

	/**
	 * reads the graphs sets and triggers the individual matchings. results are
	 * stored in the distancematrix[][]
	 */

	
	private String toMatrixWithGraphId() {
		String s = new String();
		
		/*
		 * with file names
		 */
		s += "\t";
		for(int i = 0; i < this.source.size(); i++) {
			s += ((Graph)(this.source.get(i))).getClassId() + ((Graph)(this.source.get(i))).getId() + "\t";
		}
		s += "\n";
		
		for (int i = 0; i < this.target.size(); i++){
			s += ((Graph)(this.source.get(i))).getClassId() + ((Graph)(this.target.get(i))).getId() + "\t";
			for (int j = 0; j < this.source.size(); j++)
		   		s += distanceMatrix[i][j] + "\t";		   
			s += "\n";
		}
		return s;
	}

	private void toMatrixFile(String fileName) throws IOException{
		FileWriter output = new FileWriter(fileName);
		output.write(this.toMatrixWithGraphId());
		output.flush();
		output.close();
	}

	/**
	 * the main method -- where it all begins
	 * 
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
	
		
		MatrixGenerator mgen = new MatrixGenerator();
		Munkres munkres = new Munkres();
		MunkresRec munkresRec = new MunkresRec();
		mgen.setMunkres(munkresRec);
		
		Constants.timer = new Timer();	
		
		Constants.timeconstraint = 30000;  // time limit: 30 seconds
		Constants.costFunction = new GRECCostFunction1();
		Constants.edgeHandler = new UniversalEdgeHandler(); 
		
		XMLParser xmlParser = new XMLParser();

		
	    Graph g2,g1;
		g1=xmlParser.parseGXL(args[0]);
		g2=xmlParser.parseGXL(args[1]);
		
		long start = System.nanoTime();
		double[][] matrix = mgen.getMatrix(g1, g2);
		munkres.setGraphs(g1, g2);
		double distance = munkres.getCosts(matrix);
		long end= System.nanoTime();
		
		long executionTime_NanoSeconds = end - start;
		
		double executionTime_seconds = (double)executionTime_NanoSeconds / (double)1000000000;
		
        PrintStream ps = new PrintStream(args[2]);
		
		ps.println("Opt_sol;time;node_match");
		ps.print(false+";");
		ps.print(new DecimalFormat("##.########").format(executionTime_seconds)+";");
			
		ps.println(munkres.getBestEditPath().g2Indices1(g1));
		ps.close();
			
		System.exit(0);
		
		
	}
}
