package algorithms;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Properties;
import java.util.Timer;

import util.UnDirectedEdgeHandler;
import util.EditPath;
import util.GRECCostFunction;
import util.Graph;
import util.GraphCollection;
import util.ICostFunction;
import util.IEdgeHandler;
import util.LetterCostFunction;
import util.MutagenCostFunction;
import util.PairGraphDistance;
import util.ProteinCostFunction;
import xml.XMLParser;



public class OneNNClassifierRR {

	/**
	 * @param args
	 */

	/*
	 * source is the test set
	 * target is the raining set
	 * e.g.
	 * 	source=data/Letter/MED/test.cxl
	 * 	target=data/Letter/MED/train.cxl
	 */
	GraphCollection source; // source is the test set
	private GraphCollection target; // target is the raining set
	private Graph sourceGraph, targetGraph;
	public String propFileName;
	public boolean debug;
	static private int PLAIN_GED=0;
	static private int ASTAR_GED=1;
	static private int MUNKRES_GED=2;
	static private int BEAM_GED=3;
	static private int LowerUpper_GED=4;
	static private int DepthFirstLowerUpper_GED=5;
	private static int Hausdorff_GED=6;
	private static int LapJonker=7;
	static private int OneTree_GED=8;
	private int method; // to choose between PLAIN_GED, ASTAR_GED, MNKRES_GED and BEAM_GED methods
	public String methodName; // to save the name of the method (algorithm) so as to be used in the string file
	// An Array list of GraphEditPairs where each pair equals (graph, cost)
	public ArrayList<PairGraphDistance> pairGraphDistances; 

	double alpha;
	double edgeCosts;
	double nodeCosts;
	int counter;
	double totalTime;
	// variables that help to detect the time taken to classify a graph
	public double start; // starting time
	public double end; // ending time
	public int openSize;  // for the beam search
	private String trainingcxl;
	private String testcxl;
	private String outputfilename;
	private String methods;
	private PrintStream ps;
	private GraphCollection training;
	private GraphCollection test;
	private int trainingsize;
	private int testsize;
	private String alltimeconstraint;
	private String allmemoryconstraints;
	private String[] tabmethods;
	private String[] tabtimeconstraint;
	private String[] tabmemoryconstraint;
	private ArrayList<Integer> ListOfMethods;
	private MatrixGenerator mgen;
	private Munkres munkres;
	private MunkresRec munkresRec;


	
	public OneNNClassifierRR(String trainingcxl, String testcxl,
			String outputfilename, String methods, String alltimeconstraint, String allmemoryconstraints) throws Exception {
		// TODO Auto-generated constructor stub

		this.trainingcxl=trainingcxl;
		this.testcxl=testcxl;
		this.outputfilename = outputfilename;
		this.methods = methods;
		this.alltimeconstraint=alltimeconstraint;
		this.allmemoryconstraints=allmemoryconstraints;


		tabtimeconstraint = alltimeconstraint.split(",");
		tabmemoryconstraint = allmemoryconstraints.split(",");
		Constants.timer = new Timer();	
		init();
		for(int i=0;i<tabmemoryconstraint.length;i++){
			for(int j=0;j<tabtimeconstraint.length;j++){
				Constants.timeconstraint=Long.parseLong(tabtimeconstraint[j]);
				Constants.memoryconstraint=Long.parseLong(tabmemoryconstraint[i]);
				LaunchOnAllMethods(methods);
			}	
		}


		Constants.timer.cancel();
		ps.close();
	}


	private void LaunchOnAllMethods(String medothds) throws FileNotFoundException {
		// TODO Auto-generated method stub

		ArrayList<ArrayList<Integer>> Param = new ArrayList<ArrayList<Integer>>();
		ListOfMethods = new ArrayList<Integer>();
		String[] toto = medothds.split(",");
		for(int i=0;i<toto.length;i++){
			String s = toto[i];
			int val = Integer.parseInt(s);
			ListOfMethods.add(val);
			ArrayList<Integer> params = new ArrayList<Integer>();
			Param.add(params);
			if(val==this.ASTAR_GED){
				String s1 = toto[i+1];
				int val1 = Integer.parseInt(s1);
				i=i+1;
				params.add(val1);

			}
			if(val==this.BEAM_GED){
				String s1 = toto[i+1];
				int val1 = Integer.parseInt(s1);
				i=i+1;
				params.add(val1);
			}

			if(val==this.LowerUpper_GED){
				String s1 = toto[i+1];
				int val1 = Integer.parseInt(s1);
				i=i+1;
				params.add(val1);
			}
			if(val==this.DepthFirstLowerUpper_GED){
				String s1 = toto[i+1];
				int val1 = Integer.parseInt(s1);
				i=i+1;
				params.add(val1);
				String s2 = toto[i+1];
				int val2 = Integer.parseInt(s2);
				i=i+1;
				params.add(val2);

			}
			
			if(val==this.OneTree_GED){
				String s1 = toto[i+1];
				int val1 = Integer.parseInt(s1);
				i=i+1;
				params.add(val1);
				String s2 = toto[i+1];
				int val2 = Integer.parseInt(s2);
				i=i+1;
				params.add(val2);

			}
		}
		for(int i=0;i<ListOfMethods.size();i++){
			int method = ListOfMethods.get(i);
			ArrayList<Integer> params = Param.get(i);
			RunOneMethod(method,params);

		}

	}


	private void RunOneMethod(Integer method,ArrayList<Integer> ListOfParameters) throws FileNotFoundException {
		// TODO Auto-generated method stub
		



		if(method==this.MUNKRES_GED){
			mgen = new MatrixGenerator();
			munkres = new Munkres();
			munkresRec = new MunkresRec();
			mgen.setMunkres(munkresRec);

			RunMunkresGED();
		}	


		if(method==this.DepthFirstLowerUpper_GED){


			//for(Integer param:ListOfParameters){
			//System.out.println("Param="+param);
			RunDepthFirstGraphEditDistance(ListOfParameters.get(0),ListOfParameters.get(1));			
			//}

		}
		
		if(method==this.OneTree_GED){


			//for(Integer param:ListOfParameters){
			//System.out.println("Param="+param);
			RunOneTreeDF(ListOfParameters.get(0),ListOfParameters.get(1));			
			//}

		}





	}






	private void RunDepthFirstGraphEditDistance(Integer param, Integer param2) {
		// TODO Auto-generated method stub
		double accuracy=0;
		double nbunfeasible=0;
		double nbtimeout=0;
		double nboptimal=0;
		long start = System.currentTimeMillis();
		for(int i=0;i<this.testsize;i++){
			Graph g1 = (Graph) this.test.get(i);
			System.out.println("Progression:"+(i/(double)testsize));
			this.pairGraphDistances = new ArrayList<PairGraphDistance>();
			for(int j=0;j<this.trainingsize;j++){
				Graph g2 = (Graph) this.training.get(j);
				GEDDFS GED = new GEDDFS(g1,g2,Constants.costFunction,Constants.edgeHandler,param,param2,false,false);

				double distance = GED.getBestEditpath().getTotalCosts();
				this.pairGraphDistances.add(new PairGraphDistance(g2,distance));
				if(GED.isTimeconstraintover()==true){

					nbtimeout++;
				}

				if(GED.isSolutionoptimal()==true){

					nboptimal++;
				}





			}
			Collections.sort(pairGraphDistances);
			PairGraphDistance bestgraph = pairGraphDistances.get(0);
			String bestclass = bestgraph.getGraph().getClassId();
			String classref = g1.getClassId();
			if(bestclass.equals(classref)==true){
				accuracy++;
			}
		}
		accuracy = accuracy/(double)testsize;
		long end = System.currentTimeMillis();
		long delay = end-start;
		double avgdelay = (end-start)/(double)testsize;
		ps.print("GEDDepthFirst;");
		ps.print(param+"_"+param2+";");
		ps.print(Constants.timeconstraint+";");
		ps.print(Constants.memoryconstraint+";");
		ps.print(accuracy+";");
		ps.print(avgdelay+";");
		ps.print(delay+";");
		ps.print(nbunfeasible+";");
		ps.print((nbunfeasible/(trainingsize*testsize))+";");
		ps.print(nbtimeout+";");
		ps.print((nbtimeout/(trainingsize*testsize))+";");
		ps.print(nboptimal+";");
		ps.print((nboptimal/(trainingsize*testsize))+";");
		ps.print(trainingsize+";");
		ps.println(testsize);
	}
	
	
	private void RunOneTreeDF(Integer param, Integer param2) {
		// TODO Auto-generated method stub
		double accuracy=0;
		double nbunfeasible=0;
		double nbtimeout=0;
		double nboptimal=0;
		long start = System.currentTimeMillis();
		for(int i=0;i<this.testsize;i++){
			Graph g1 = (Graph) this.test.get(i);
			System.out.println("Progression:"+(i/(double)testsize));
			this.pairGraphDistances = new ArrayList<PairGraphDistance>();
			double prevUpperBound=Double.MAX_VALUE;
			EditPath prevEditPath=null;
			String bestClass="";
			double isFirstComparision=0;
			for(int j=0;j<this.trainingsize;j++){
				Graph g2 = (Graph) this.training.get(j);
				OneTreeDF GED = new OneTreeDF(g1,g2,Constants.costFunction,Constants.edgeHandler,param,param2,prevUpperBound,prevEditPath,bestClass,isFirstComparision,false,false);

				double distance = GED.getBestEditpath().getTotalCosts();
				if(prevUpperBound>distance) {
					prevUpperBound=distance;	
					prevEditPath=GED.getBestEditpath();
					
					bestClass=g2.getClassId();
					isFirstComparision=1;
				}
				
				this.pairGraphDistances.add(new PairGraphDistance(g2,distance));
				if(GED.isTimeconstraintover()==true){

					nbtimeout++;
				}

				if(GED.isSolutionoptimal()==true){

					nboptimal++;
				}





			}
			Collections.sort(pairGraphDistances);
			PairGraphDistance bestgraph = pairGraphDistances.get(0);
			String bestclass = bestgraph.getGraph().getClassId();
			String classref = g1.getClassId();
			if(bestclass.equals(classref)==true){
				accuracy++;
			}
		}
		accuracy = accuracy/(double)testsize;
		long end = System.currentTimeMillis();
		long delay = end-start;
		double avgdelay = (end-start)/(double)testsize;
		ps.print("GEDDepthFirst;");
		ps.print(param+"_"+param2+";");
		ps.print(Constants.timeconstraint+";");
		ps.print(Constants.memoryconstraint+";");
		ps.print(accuracy+";");
		ps.print(avgdelay+";");
		ps.print(delay+";");
		ps.print(nbunfeasible+";");
		ps.print((nbunfeasible/(trainingsize*testsize))+";");
		ps.print(nbtimeout+";");
		ps.print((nbtimeout/(trainingsize*testsize))+";");
		ps.print(nboptimal+";");
		ps.print((nboptimal/(trainingsize*testsize))+";");
		ps.print(trainingsize+";");
		ps.println(testsize);
	}


	private void RunMunkresGED() {
		// TODO Auto-generated method stub

		double accuracy=0;
		double nbunfeasible=0;
		double nbtimeout=0;
		double nboptimal=0;
		long start = System.currentTimeMillis();
		for(int i=0;i<this.testsize;i++){
			Graph g1 = (Graph) this.test.get(i);
			System.out.println("Progression:"+(i/(double)testsize));
			this.pairGraphDistances = new ArrayList<PairGraphDistance>();
			for(int j=0;j<this.trainingsize;j++){

				Graph g2 = (Graph) this.training.get(j);
				double[][] matrix = mgen.getMatrix(g1, g2);
				munkres.setGraphs(g1, g2);
				double distance = munkres.getCosts(matrix);	
				
				this.pairGraphDistances.add(new PairGraphDistance(g2,distance));
				if(munkres.isTimeconstraintover()==true){
					nbunfeasible++;
					nbtimeout++;
				}

			}
			Collections.sort(pairGraphDistances);
			PairGraphDistance bestgraph = pairGraphDistances.get(0);
			String bestclass = bestgraph.getGraph().getClassId();
			String classref = g1.getClassId();
			if(bestclass.equals(classref)==true){
				accuracy++;
			}
			pairGraphDistances.clear();
			pairGraphDistances=null;
		}
		accuracy = accuracy/(double)testsize;
		long end = System.currentTimeMillis();
		long delay = end-start;
		double avgdelay = (end-start)/(double)testsize;
		ps.print("GEDMunkres;");
		ps.print("rien"+";");
		ps.print(Constants.timeconstraint+";");
		ps.print(Constants.memoryconstraint+";");
		ps.print(accuracy+";");
		ps.print(avgdelay+";");
		ps.print(delay+";");
		ps.print(nbunfeasible+";");
		ps.print((nbunfeasible/(trainingsize*testsize))+";");
		ps.print(nbtimeout+";");
		ps.print((nbtimeout/(trainingsize*testsize))+";");
		ps.print(nboptimal+";");
		ps.print((nboptimal/(trainingsize*testsize))+";");
		ps.print(trainingsize+";");
		ps.println(testsize);



	}



	/*private void RunPlainGED()  {
		// TODO Auto-generated method stub
		double accuracy=0;
		long start = System.currentTimeMillis();
		for(int i=0;i<this.testsize;i++){
			Graph g1 = (Graph) this.test.get(i);
			System.out.println("Progression:"+(i/(double)testsize));
			this.pairGraphDistances = new ArrayList<PairGraphDistance>();
			for(int j=0;j<this.trainingsize;j++){
				Graph g2 = (Graph) this.training.get(j);
				GraphEditDistance GED = new GraphEditDistance(g1,g2,Constants.costFunction,Constants.edgeHandler,false);
				double distance = GED.getBestEditpath().getTotalCosts();
				this.pairGraphDistances.add(new PairGraphDistance(g2,distance));
				System.gc();
			}
			Collections.sort(pairGraphDistances);
			PairGraphDistance bestgraph = pairGraphDistances.get(0);
			String bestclass = bestgraph.getGraph().getClassId();
			String classref = g1.getClassId();
			if(bestclass.equals(classref)==true){
				accuracy++;
			}
		}
		accuracy = accuracy/(double)testsize;
		long end = System.currentTimeMillis();
		long delay = end-start;
		double avgdelay = (end-start)/(double)testsize;
		ps.print("GEDA*PLAIN;");
		ps.print("rien"+";");
		ps.print(Constants.timeconstraint+";");
		ps.print(Constants.memoryconstraint+";");
		ps.print(accuracy+";");
		ps.print(avgdelay+";");
		ps.println(delay);
	}*/








	private void init() throws Exception {
		// TODO Auto-generated method stub


		// set the appropriate edgehandler (directed or undirected)
		Constants.edgeHandler = new UnDirectedEdgeHandler();//UnDirectedEdgeHandler(); 
		ps = new PrintStream(this.outputfilename);
		ps.println("Method;Param;Time Constraints;Mmory Constraints;Accuracy;Average response time;Total response time;# of unfeasible solution; Average # of unfeasible solution;# of time out cases; Average # of time out cases;# of optimal solution; Average # of optimal solution;  Training size; Test Size ");
		//ps.println(""+this.graphTab.length+";;;;;;;;;;;;;;;");
		LoadGraph();	
	}


	private void LoadGraph() throws Exception {
		// TODO Auto-generated method stub

		XMLParser xmlParser = new XMLParser();
		// Getting the path of the Dataset
		xmlParser.setGraphPath(new File(this.trainingcxl).getParent()+"/");
		// parsing the training set
		this.training = xmlParser.parseCXL(this.trainingcxl);
		// parsing the test set
		this.test = xmlParser.parseCXL(this.testcxl);

		trainingsize = training.size();
		testsize = test.size();

	}


	private void printPairs() {
		// TODO Auto-generated method stub
		for (int i=0; i<pairGraphDistances.size(); i++) {
			PairGraphDistance pr = pairGraphDistances.get(i);
			Graph g = pr.getGraph();
			double dist = pr.getdist();
			System.out.println(g.getId() + ", " + dist);
		}
	}

	//  Comparing the class of the test and the nearest neighbor's class
	private void CompareClasses(Graph src,
			ArrayList<PairGraphDistance> pairGraphDistances) 
	{
		// TODO Auto-generated method stub

		Collections.sort(pairGraphDistances);
		Graph tgt = pairGraphDistances.get(0).getGraph();
		if(tgt.getClassId().equals(src.getClassId()))
		{
			if (debug==true) System.out.println(""+src.getId()+"and "+tgt.getId()+" have the same class which is : "+tgt.getClassId());
			this.counter++;
		}
		else
		{
			if (debug==true) System.out.println(""+src.getId()+"and "+tgt.getId()+" doesn't have the same class");

		}

	} 

	// A funtion that computes the accuracy of classifying graphs 
	double getAccuracy(String saveFile) throws FileNotFoundException {
		// TODO Auto-generated method stub
		PrintStream ps =  new PrintStream(saveFile+methodName+"Accuracy.csv") ;
		double precision = (double)((double)counter/(double) source.size());
		//ps.println("precision");
		ps.println(precision);
		ps.close();
		return precision;

	}

	// A funtion that returns the number of the well-classified graphs 
	int getCorrectlyClassifiedInstances(String saveFile) throws FileNotFoundException {
		// TODO Auto-generated method stub
		PrintStream ps =  new PrintStream(saveFile+methodName+"CorrectlyClassifiedInstances.csv");
		//ps.println("CorrectlyClassifiedInstances");
		ps.println(counter);
		ps.close();
		return counter;

	}

	double getTotalClassificationTime(String saveFile) throws FileNotFoundException {
		// TODO Auto-generated method stub
		PrintStream ps =  new PrintStream(saveFile+methodName+"totalClassificationTime.csv");
		//ps.println("TotalClassificationTime");
		ps.println(totalTime);
		ps.close();
		return totalTime;

	}

	public static void main(String[] args) throws Exception {
		System.out.println("Usage : \n 1 : CXL file of the training set  \n 2 : CXL file of the test set  \n 3 : Output csv filename  \n 4 : Database ID {0=GREC, 1=MUTA, 2=LETTER, 4=Protein} \n 5: time limit in milliseconds \n 6: method id : {2=BP, 5=DF, 8=OntreeDF} DF and OneTree have 2 parameters the lower and upper bound methods see below for more details, \n 7: memory limit in Mega bytes");
		System.out.println("\tMethod=5  GED deep first search (Branch and Bound). !!! Two parameters needed !!! Exemple 5,1,0");
		System.out.println("\t\tParam1= Lower Bound Heuristic");
		System.out.println("\t\tParam1=0  NoHeuristic");
		System.out.println("\t\tParam1=1  NoAssigmentHeuristic");
		System.out.println("\t\tParam1=2  MunkresAssigmentHeuristic");
		System.out.println("\t\tParam1=3  LAPAssigmentHeuristic");
		System.out.println("\t\tParam1=4  HausdorffEDHeuristic");
		System.out.println("\t\tParam2= Upper Bound Heuristic");
		System.out.println("\t\tParam2=0  Munkres Heuristic (BP)");
		System.out.println("\t\tParam2=1  Dummy Upper Bound");
		
		System.out.println("\tMethod=8  OneTree DF (Branch and Bound). !!! Two parameters needed !!! Example 8,1,0");
		System.out.println("\t\tParam1= Lower Bound Heuristic");
		System.out.println("\t\tParam1=0  NoHeuristic");
		System.out.println("\t\tParam1=1  NoAssigmentHeuristic");
		System.out.println("\t\tParam1=2  MunkresAssigmentHeuristic");
		System.out.println("\t\tParam1=3  LAPAssigmentHeuristic");
		System.out.println("\t\tParam1=4  HausdorffEDHeuristic");
		System.out.println("\t\tParam2= Upper Bound Heuristic");
		System.out.println("\t\tParam2=0  Munkres Heuristic (BP)");
		System.out.println("\t\tParam2=1  Dummy Upper Bound");
		
		
		String trainingcxl=args[0];
		String testcxl=args[1];
		String Outputfilename=args[2];
		int databaseid=Integer.parseInt(args[3]);
		String alltimeconstraint=args[4];
		String methods=args[5];
		String allmemoryconstraints= args[6];


		System.out.println("Arguments");
		for(String arg : args){
			System.out.println(arg);
		}
		System.out.println("");

		if(databaseid==0){
			GRECtest(trainingcxl,testcxl,Outputfilename,methods,alltimeconstraint,allmemoryconstraints);
		}
		if(databaseid==1){
			MUTAtest(trainingcxl,testcxl,Outputfilename,methods,alltimeconstraint,allmemoryconstraints);

		}

		if(databaseid==2){
			LETTERtest(trainingcxl,testcxl,Outputfilename,methods,alltimeconstraint,allmemoryconstraints);

		}


		if(databaseid==4){
			Proteintest(trainingcxl,testcxl,Outputfilename,methods,alltimeconstraint,allmemoryconstraints);

		}

		




	}



	private static void Proteintest(String trainingcxl, String testcxl,
			String outputfilename, String methods, String alltimeconstraint, String allmemoryconstraints) throws Exception {

		double n=11;
		double e=1;
		double a=0.75;
		Constants.costFunction = (ICostFunction)new ProteinCostFunction(n,e,a); 
		new OneNNClassifierRR(trainingcxl,testcxl,outputfilename,methods,alltimeconstraint,allmemoryconstraints);


	}


	private static void GRECtest(String trainingcxl, String testcxl,
			String outputfilename, String methods, String alltimeconstraint, String allmemoryconstraints) throws Exception {
		// TODO Auto-generated method stub
		double n=90;
		double e=15;
		double a=0.5;
		Constants.costFunction = (ICostFunction)new GRECCostFunction(n,e,a); 
		new OneNNClassifierRR(trainingcxl,testcxl,outputfilename,methods,alltimeconstraint,allmemoryconstraints);
	}

	private static void MUTAtest(String trainingcxl, String testcxl,
			String outputfilename, String methods, String alltimeconstraint, String allmemoryconstraints) throws Exception {
		// TODO Auto-generated method stub
		double n=11;
		double e=1.1;
		double a=0.25;
		Constants.costFunction = (ICostFunction)new MutagenCostFunction(n,e,a); 
		new OneNNClassifierRR(trainingcxl,testcxl,outputfilename,methods,alltimeconstraint,allmemoryconstraints);
	}

	private static void LETTERtest(String trainingcxl, String testcxl,
			String outputfilename, String methods, String alltimeconstraint, String allmemoryconstraints) throws Exception {
		// TODO Auto-generated method stub
		double n=0.3;
		double e=0.5;
		double a=0.75;
		Constants.costFunction = (ICostFunction)new LetterCostFunction(n,e,a); 
		new OneNNClassifierRR(trainingcxl,testcxl,outputfilename,methods,alltimeconstraint,allmemoryconstraints);
	}



}
